use crate::{
    minimize::{ExitCondition, OptimizationResult},
    objective_function::ObjectiveFun,
    traits::XVar,
};
use std::fmt::Debug;

use super::line_search::LineSearch;

pub trait SteepestDescentPrime {
    fn scale(&self, factor: f64) -> Self;
    fn inner_product(&self, rhs: &Self) -> f64;
}

impl SteepestDescentPrime for f64 {
    fn scale(&self, factor: f64) -> Self {
        self * factor
    }

    fn inner_product(&self, rhs: &Self) -> f64 {
        self * rhs
    }
}

impl SteepestDescentPrime for Vec<f64> {
    fn scale(&self, factor: f64) -> Self {
        self.iter().map(|val| val * factor).collect()
    }

    fn inner_product(&self, rhs: &Self) -> f64 {
        self.iter()
            .zip(rhs)
            .fold(0.0, |acc, (lhs, rhs)| acc + lhs * rhs)
    }
}
pub fn steepest_descent<T: XVar<E> + Clone, E: Debug + SteepestDescentPrime>(
    fun: &dyn ObjectiveFun<T, E>,
    x0: &T,
    max_iters: usize,
    tolerance: f64,
    line_search: &LineSearch,
) -> OptimizationResult<T> {
    // Make a mutable copy of x0 to work with
    let mut xs = x0.clone();

    // Perform the iteration
    let mut f_iminus1 = f64::INFINITY;
    let mut f = 0.0;
    let mut i = 0;
    for _ in 0..max_iters {
        let direction = &fun.prime(&xs).scale(-1.0);
        let learning_rate = line_search.get_learning_rate(fun, &xs, direction);
        xs = xs.update(learning_rate, direction);
        f = fun.eval(&xs);
        if (f - f_iminus1).abs() < tolerance {
            break;
        } else {
            f_iminus1 = f;
        }
        i += 1;
    }

    let exit_con = if i == max_iters {
        ExitCondition::MaxIter
    } else {
        ExitCondition::Converged
    };
    OptimizationResult {
        best_xs: xs,
        best_fun_val: f,
        exit_con,
        iters: i,
    }
}

#[cfg(test)]
mod test {
    use crate::objective_function::Fun;

    use super::*;

    #[test]
    pub fn simple_steepest_descent_test() {
        let fun = Box::new(|xs: &Vec<f64>| xs.iter().fold(0.0, |acc, x| acc + x.powi(2)));
        let prime = Box::new(|xs: &Vec<f64>| xs.iter().map(|x| 2.0 * x).collect());

        let obj = Fun::new(fun, prime);
        let line_searches = vec![
            LineSearch::ConstAlpha {
                learning_rate: 0.25,
            },
            LineSearch::BackTrack {
                gamma: 0.5,
                c: 0.001,
            },
        ];
        for line_search in line_searches {
            let res = steepest_descent(&obj, &vec![20.0, 20.0], 1000, 1e-12, &line_search);

            if let ExitCondition::MaxIter = res.exit_con {
                panic!("Failed to converge to minima");
            }
            println!(
                "{:?} on iteration {} has value:\n{}",
                res.best_xs, res.iters, res.best_fun_val
            );
            assert!(res.best_fun_val < 1e-8);
        }
    }

    #[test]
    pub fn basic_beale_test() {
        let fun = Box::new(|x: &Vec<f64>| {
            (1.5 - x[0] + x[0] * x[1]).powi(2)
                + (2.25 - x[0] + x[0] * x[1].powi(2)).powi(2)
                + (2.625 - x[0] + x[0] * x[1].powi(3)).powi(2)
        });
        let prime = Box::new(|x: &Vec<f64>| {
            vec![
                2.0 * (1.5 - x[0] + x[0] * x[1]) * (x[1] - 1.0)
                    + 2.0 * (2.25 - x[0] + x[0] * x[1].powi(2)) * (x[1].powi(2) - 1.0)
                    + 2.0 * (2.625 - x[0] + x[0] * x[1].powi(3)) * (x[1].powi(3) - 1.0),
                2.0 * (1.5 - x[0] + x[0] * x[1]) * (x[0])
                    + 2.0 * (2.25 - x[0] + x[0] * x[1].powi(2)) * (2.0 * x[0] * x[1])
                    + 2.0 * (2.625 - x[0] + x[0] * x[1].powi(3)) * (3.0 * x[0] * x[1].powi(3)),
            ]
        });
        let obj = Fun::new(fun, prime);
        let line_search = LineSearch::BackTrack {
            gamma: 0.9,
            c: 0.01,
        };
        let res = steepest_descent(&obj, &vec![3.1, 0.5], 10000, 1e-12, &line_search);
        println!(
            "Best val is {:?} for xs {:?} in {} iterations",
            res.best_fun_val, res.best_xs, res.iters,
        );
        assert!(res.best_fun_val < 1e-7);
    }
}
