use std::fmt::Debug;

/// Trait defining the data structure that must be implemented for the independent variables used
/// in the objective function. The generic type denotes the type of the prime of that variable
/// NOTE: This trait also defines some functions that are required to operate on the prime data
/// type. It should be noted that we are unable to just require T to implement Mul<f64> or Add<f64>
/// Bbecause then we wouldn't be able to implement XVar for plain Vec types which seems
/// inconvenient
pub trait XVar<T>: Clone + Debug {
    /// Update the current Xvariable based on the prime
    fn update(&self, alpha: f64, prime: &T) -> Self;
}

/// Implementation of XVar for an f64 type
impl XVar<f64> for f64 {
    fn update(&self, alpha: f64, prime: &f64) -> Self {
        self + alpha * prime
    }
}

/// Implementation of XVar for a Vec<f64> type
impl XVar<Vec<f64>> for Vec<f64> {
    fn update(&self, alpha: f64, prime: &Vec<f64>) -> Self {
        self.iter()
            .zip(prime.iter())
            .map(|(x, xprime)| x + alpha * xprime)
            .collect()
    }
}
